/*
								+-----------------------------------+
								|                                   |
								|     ***   Bitmap writer   ***     |
								|                                   |
								|    Design by Lukas Polok 2007     |
								|    Implemented by Jiri Hanus      |
								|                                   |
								|           BmpOutput.cpp           |
								|                                   |
								+-----------------------------------+
*/

#include <string.h>
#include <stdio.h>
#include "BmpOutput.h"

typedef short WORD;
typedef int DWORD;

#define BI_RGB       0
#define BI_RLE8      1
#define BI_RLE4      2
#define BI_BITFIELDS 3

#pragma pack(2)

typedef struct {
    WORD           bfType;
    DWORD          bfSize;
    WORD           bfReserved1;
    WORD           bfReserved2;
    DWORD          bfOffBits;
} BITMAPFILEHEADER;

typedef struct {
    DWORD          biSize;
    DWORD          biWidth;
    DWORD          biHeight;
    WORD           biPlanes;
    WORD           biBitCount;
    DWORD          biCompression;
    DWORD          biSizeImage;
    DWORD          biXPelsPerMeter;
    DWORD          biYPelsPerMeter;
    DWORD          biClrUsed;
    DWORD          biClrImportant;
} BITMAPINFOHEADER;

#pragma pack()

/*
 *	bool Save_BMP(const char *p_s_file, int n_width, int n_height,
 *		const unsigned int *p_buffer)
 *		- write RGB (windows) bitmap image to file p_s_file
 *		- returns true on success, false on failure
 *
 *		- zadn:
 *			- tahle funkce m uloit RGB bitmapu o rozmrech n_width x n_height do souboru
 *			  s nzvem p_s_file (jednodue pepsat pokud existuje)
 *			- p_buffer obsahuje jednotliv pixely bitmapy, jak jdou po dcch, zdola nahoru.
 *			- kad pixel m hodnotu (r << 16) + (g << 8) + b, kde r, g, b jsou unsigned char
 *			- vrac se true pi spchu, false pi chyb dn hlen se nevypisuje)
 */
bool Save_BMP(const char *p_s_file, int n_width, int n_height, const unsigned int *p_buffer)
{
	FILE *p_fw;
	if(!(p_fw = fopen(p_s_file, "wb")))
		return false;

	BITMAPINFOHEADER t_info_header;
	memset(&t_info_header, 0, sizeof(BITMAPINFOHEADER));
	t_info_header.biSize = sizeof(BITMAPINFOHEADER);
	t_info_header.biWidth = n_width;
	t_info_header.biHeight = n_height;
	t_info_header.biPlanes = 1;
	t_info_header.biBitCount = 24;
	t_info_header.biCompression = BI_RGB;
	t_info_header.biSizeImage = n_width * n_height * 3;
	t_info_header.biXPelsPerMeter = 96;
	t_info_header.biYPelsPerMeter = 96;

	BITMAPFILEHEADER t_file_header;
	memset(&t_file_header, 0, sizeof(BITMAPFILEHEADER));
	t_file_header.bfType = ('B' | ('M' << 8));
	t_file_header.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	t_file_header.bfSize = t_file_header.bfOffBits + t_info_header.biSizeImage;
	// create header

	if(fwrite(&t_file_header, sizeof(BITMAPFILEHEADER), 1, p_fw) != 1 ||
	   fwrite(&t_info_header, sizeof(BITMAPINFOHEADER), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write header

	int n_scanline_padd = (3 * n_width) & 3;
	// length of scanline padding (padd to 4-byte multiples)

	for(const unsigned int *p_raster = p_buffer + ((n_width - 1) * n_height);
	   p_raster >= p_buffer; p_raster -= 2 * n_width) {
		for(const unsigned int *p_scanline_end = p_raster + n_width;
		   p_raster != p_scanline_end; ++ p_raster) {
			unsigned char p_rgb[] = {
				(*p_raster >> 16) & 0xff,
				(*p_raster >> 8) & 0xff,
				*p_raster & 0xff
			};
			if(fwrite(p_rgb, sizeof(unsigned char), 3, p_fw) != 3) {
				fclose(p_fw);
				return false;
			}
		}
		// write scanline

		static const unsigned char p_dummy[3] = {0};
		if(n_scanline_padd && fwrite(p_dummy, n_scanline_padd, 1, p_fw) != 1) {
			fclose(p_fw);
			return false;
		}
		// write padding
	}
	// write bitmap (upside-down)

	fclose(p_fw);
	return true;
}

/*
 *		-end-of-file-
 */
