/*
								+----------------------------------+
								|                                  |
								| *** Varcoeff error diffusion *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|             Dither.h             |
								|                                  |
								+----------------------------------+
*/

/*
 *	based on Victor Ostromoukhov siggraph'01 paper
 *	"A Simple and Efficient Error-Diffusion Algorithm"
 */

#ifndef __VARIABLE_COEFF_ERROR_DIFFUSION_INCLUDED
#define __VARIABLE_COEFF_ERROR_DIFFUSION_INCLUDED

/*
 *	class CVarcoeffED
 *		- variable coefficients error diffusion class
 */
class CVarcoeffED {
protected:
	struct TDitherCoeffs {
		int n_r, n_dl, n_d;
		int n_sum;
	};

	static const TDitherCoeffs p_coefs_tab[256];

public:
	/*
	 *	static Vector3f v_Error(unsigned char n_original, float f_diff)
	 *		- n_original is output color, f_diff is difference to ideal color
	 *		- returns trinity of error diffusion terms to be distributed right, down-left, down
	 */
	static Vector3f v_Error(unsigned char n_original, float f_diff)
	{
		TDitherCoeffs t_coeffs = p_coefs_tab[n_original];

		float f_term_r = (t_coeffs.n_r * f_diff) / t_coeffs.n_sum;
		float f_term_dl = (t_coeffs.n_dl * f_diff) / t_coeffs.n_sum;
		float f_term_d = f_diff - (f_term_r + f_term_dl);
		return Vector3f(f_term_r, f_term_dl, f_term_d);
	}
};

#endif //__VARIABLE_COEFF_ERROR_DIFFUSION_INCLUDED
