/*
								+----------------------------------+
								|                                  |
								| *** berLame material system *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|            Material.h            |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code violation at 2007-09-27
 *
 *	removed all but material lexer, using this to parse config file
 *	might be buggy (excerpted from not very well debugged code), do not copy
 *
 *	2007-10-19
 *
 *	added convenience inline float CMaterialLexer::TToken::f_Value_AllowCast() const and
 *	inline bool CMaterialLexer::Get_NextToken(int n_accept_type, int n_accept_type2)
 *
 */

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#include <crtdbg.h>
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#endif
//#include <GL/gl.h>
//#include <GL/glext.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <algorithm>
#include "Scanner.h"
#include "Material.h"

#if defined(_MSC_VER) && !defined(_MWERKS)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 *								=== CMaterialLexer::TToken ===
 */

/*
 *	CMaterialLexer::TToken::TToken()
 *		- default constructor; sets line and column to 0 and type to token_EOF
 */
CMaterialLexer::TToken::TToken()
	:n_line(0), n_col(0), n_type(token_EOF)
{
	t_value.p_s_value = 0;
}

/*
 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type)
 *		- constructor for tokens with no data
 *		- sets line to _n_line, column to _n_col and type to _n_type
 */
CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type)
	:n_line(_n_line), n_col(_n_col), n_type(_n_type)
{
	t_value.p_s_value = 0;
}

/*
 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, int n_value)
 *		- constructor for tokens with integer data
 *		- sets line to _n_line, column to _n_col and type to _n_type
 *		- sets int value to n_value
 */
CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, int n_value)
	:n_line(_n_line), n_col(_n_col), n_type(_n_type)
{
	t_value.n_value = n_value;
}

/*
 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, float f_value)
 *		- constructor for tokens with floating point data
 *		- sets line to _n_line, column to _n_col and type to _n_type
 *		- sets float value to f_value
 */
CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, float f_value)
	:n_line(_n_line), n_col(_n_col), n_type(_n_type)
{
	t_value.f_value = f_value;
}

/*
 *	CMaterialLexer::TToken::TToken(int _n_line, int _n_col,
 *		int _n_type, const char *p_s_value)
 *		- constructor for tokens with string data
 *		- sets line to _n_line, column to _n_col and type to _n_type
 *		- sets string value to p_s_value (a new string is allocated and copied,
 *		  check wheter allocation was succesful)
 */
CMaterialLexer::TToken::TToken(int _n_line, int _n_col, int _n_type, const char *p_s_value)
	:n_line(_n_line), n_col(_n_col), n_type(_n_type)
{
	if(n_type >= token_Node && n_type <= token_CData && p_s_value) {
		if((t_value.p_s_value = new char[strlen(p_s_value) + 1]))
			strcpy(t_value.p_s_value, p_s_value);
	} else
		t_value.p_s_value = 0;
}

/*
 *	CMaterialLexer::TToken::TToken(const TToken &r_t_token)
 *		- copy-constructor
 *		- note in case token is supposed to contain string data, a new string
 *		  is allocated and copied, check wheter allocation was succesful
 */
CMaterialLexer::TToken::TToken(const TToken &r_t_token)
	:n_line(r_t_token.n_line), n_col(r_t_token.n_col), n_type(r_t_token.n_type)
{
	t_value = r_t_token.t_value;
	if(n_type >= token_Node && n_type <= token_CData && t_value.p_s_value) {
		if((t_value.p_s_value = new char[strlen(r_t_token.t_value.p_s_value) + 1]))
			strcpy(t_value.p_s_value, r_t_token.t_value.p_s_value);
	}
}

/*
 *	CMaterialLexer::TToken::~TToken()
 *		- default destructor; deallocates string data in case it was present
 */
CMaterialLexer::TToken::~TToken()
{
	if(n_type >= token_Node && n_type <= token_CData && t_value.p_s_value)
		delete[] t_value.p_s_value;
}

/*
 *	bool CMaterialLexer::TToken::operator =(const TToken &r_t_token)
 *		- copy-operator
 *		- returns true on success, false in case there was
 *		  not enough memory for string data
 */
bool CMaterialLexer::TToken::operator =(const TToken &r_t_token)
{
	if(n_type >= token_Node && n_type <= token_CData && t_value.p_s_value)
		delete[] t_value.p_s_value;
	// delete previous string

	n_line = r_t_token.n_line;
	n_col = r_t_token.n_col;
	n_type = r_t_token.n_type;
	t_value = r_t_token.t_value;
	// copy token data

	if(n_type >= token_Node && n_type <= token_CData && t_value.p_s_value) {
		if((t_value.p_s_value = new char[strlen(r_t_token.t_value.p_s_value) + 1]))
			strcpy(t_value.p_s_value, r_t_token.t_value.p_s_value);
		else
			return false;
	}
	// copy new string

	return true;
}

#ifdef MATERIAL_LEXER_TOKEN_DUMP

/*
 *	void CMaterialLexer::TToken::Dump() const
 *		- prints token contents
 */
void CMaterialLexer::TToken::Dump() const
{
	const char *p_type_name_list[] = {
		"token_White",
		"token_Colon",
		"token_Semicolon",
		"token_LeftPar",
		"token_RightPar",
		"token_Begin",
		"token_End",
		"token_Node",
		"token_String",
		"token_CData",
		"token_Bool",
		"token_Int",
		"token_GLenum",
		"token_Float",
		"token_Pass",
		"token_Texture1D",
		"token_Texture2D",
		"token_Texture3D",
		"token_TextureCube",
		"token_URI",
		"token_Reflect",
		"token_Refract",
		"token_Custom",
		"token_DepthMap",
		"token_EOF"
	};

	if(n_type == token_Bool || n_type == token_Int) {
		printf("TToken(n_line = %d, n_col = %d, n_type = %s, n_value = %d)\n",
			n_line, n_col, p_type_name_list[n_type], t_value.n_value);
	} else if(n_type == token_Float) {
		printf("TToken(n_line = %d, n_col = %d, n_type = %s, f_value = %g)\n",
			n_line, n_col, p_type_name_list[n_type], t_value.f_value);
	} else if(n_type == token_String || n_type == token_Node || n_type == token_CData) {
		printf("TToken(n_line = %d, n_col = %d, n_type = %s, p_s_value = \'%s\')\n",
			n_line, n_col, p_type_name_list[n_type], (t_value.p_s_value)?
			t_value.p_s_value : "(null)");
	} else if(n_type == token_GLenum) {
		printf("TToken(n_line = %d, n_col = %d, n_type = %s, n_value = 0x%04x)\n",
			n_line, n_col, p_type_name_list[n_type], t_value.n_value);
	} else {
		printf("TToken(n_line = %d, n_col = %d, n_type = %s)\n",
			n_line, n_col, p_type_name_list[n_type]);
	}
}

#endif //MATERIAL_LEXER_TOKEN_DUMP

/*
 *								=== ~CMaterialLexer::TToken ===
 */

/*
 *								=== CMaterialLexer::CFileReader ===
 */

class CMaterialLexer::CFileReader {
protected:
	FILE *m_p_fr;

public:
	CFileReader(const char *p_s_filename)
		:m_p_fr(fopen(p_s_filename, "r"))
	{}

	~CFileReader()
	{
		if(m_p_fr)
			fclose(m_p_fr);
	}

	bool b_Status() const
	{
		return m_p_fr != 0;
	}

	inline bool ReadChar(int &r_n_char)
	{
		return m_p_fr && (r_n_char = fgetc(m_p_fr)) != EOF;
	}

	inline bool UnreadChar(int n_char)
	{
		return m_p_fr && ungetc(n_char, m_p_fr);
	}
};

/*
 *								=== ~CMaterialLexer::CFileReader ===
 */

/*
 *								=== CMaterialLexer::CTokenEmit ===
 */

class CMaterialLexer::CTokenEmit {
protected:
	CMaterialLexer::TToken &m_r_t_out_token;
	bool &m_r_b_validity_flag;

public:
	CTokenEmit(TToken &r_t_out_token, bool &r_b_validity_flag)
		:m_r_t_out_token(r_t_out_token), m_r_b_validity_flag(r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const = 0;
};

class CMaterialLexer::CEmitSkip : public CTokenEmit {
public:
	CEmitSkip(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 0 && p_s_buffer);

		m_r_b_validity_flag = true;
		m_r_t_out_token = TToken(n_line, n_column, token_White);
		return true;
	}
};

class CMaterialLexer::CEmitNode : public CTokenEmit {
public:
	CEmitNode(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 1);

		m_r_b_validity_flag = true;

		const struct {
			const char *p_s_name;
			int n_type;
		} p_keyword[] = {
			{"false", token_Bool},
			{"true", token_Bool},
			// hack! bool value equals to index in this array

			{"uri", token_URI},
			{"tex-2d", token_Texture2D},
			{"tex-1d", token_Texture1D},
			{"tex-3d", token_Texture3D},
			{"tex-cube", token_TextureCube},
			{"pass", token_Pass},
			{"reflect", token_Reflect},
			{"refract", token_Refract},
			{"custom", token_Custom},
			{"depth-map", token_DepthMap}
			// otherwise int value insignificant, sorted by occurence likehood
		};
		for(int i = 0; i < (int)(sizeof(p_keyword) / sizeof(p_keyword[0])); ++ i) {
			if(!strcmp(p_s_buffer, p_keyword[i].p_s_name)) {
				m_r_t_out_token = TToken(n_line, n_column,
					 p_keyword[i].n_type, i);
				return true;
			}
		}

		m_r_t_out_token = TToken(n_line, n_column, token_Node, p_s_buffer);
		return m_r_t_out_token.p_s_Value() != 0;
	}
};

class CMaterialLexer::CEmitOperator : public CTokenEmit {
public:
	CEmitOperator(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 2);

		m_r_b_validity_flag = true;
		const char *p_s_ops = ":;(){}";
		const char *p_s_thisop = strchr(p_s_ops, *p_s_buffer);
		_ASSERTE(p_s_thisop); // operators given by FSM, not neccesary to double-check
		m_r_t_out_token = TToken(n_line, n_column, token_Colon + (p_s_thisop - p_s_ops));
		return true;
	}
};

class CMaterialLexer::CEmitString : public CTokenEmit {
public:
	CEmitString(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 3);

		m_r_b_validity_flag = true;
		int n_length = strlen(p_s_buffer);
		char *p_s_string;
		if(!(p_s_string = new char[n_length - 1]))
			return false;
		// alloc buffer

		if(*p_s_buffer == '\'') {
			strncpy(p_s_string, p_s_buffer + 1, n_length - 2);
			p_s_string[n_length - 2] = 0;
			// copy 'string', it won't contain any escape sequences
		} else {
			_ASSERTE(*p_s_buffer == '\"');
			++ p_s_buffer;
			for(char *p_s_out = p_s_string;; ++ p_s_buffer, ++ p_s_out) {
				if(*p_s_buffer == '\\') {
					char n_char;
					switch(n_char = *(++ p_s_buffer)) {
					case 'n':
						*p_s_out = '\n';
						break;
					case 't':
						*p_s_out = '\t';
						break;
					case 'r':
						*p_s_out = '\r';
						break;
					case '\\':
					case '\'':
					case '\"':
						*p_s_out = n_char;
						break;
					default:
						_ASSERTE(0);
						// escape sequences given by FSM, not neccesary to double-check
						break;
					}
				} else if(*p_s_buffer == '\"') {
					*p_s_out = 0;
					break;
				} else
					*p_s_out = *p_s_buffer;
			}
			// copy "string", process escape sequences
		}

		m_r_t_out_token = TToken(n_line, n_column, token_String, p_s_string);
		delete[] p_s_string;
		return m_r_t_out_token.p_s_Value() != 0;
	}
};

class CMaterialLexer::CEmitInt : public CTokenEmit {
public:
	CEmitInt(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 4);

		m_r_b_validity_flag = true;
		m_r_t_out_token = TToken(n_line, n_column, token_Int, (int)atol(p_s_buffer));
		// (int) to avoid ambiguity
		return true;
	}
};

class CMaterialLexer::CEmitFloat : public CTokenEmit {
public:
	CEmitFloat(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 5);

		m_r_b_validity_flag = true;
		m_r_t_out_token = TToken(n_line, n_column, token_Float, (float)atof(p_s_buffer));
		return true;
	}
};

class CMaterialLexer::CEmitGLenum : public CTokenEmit {
public:
	CEmitGLenum(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 6);

		m_r_b_validity_flag = true;
		int n_enum;
		if((n_enum = CGLenumTable::n_Value(p_s_buffer)) < 0)
			return false;
		m_r_t_out_token = TToken(n_line, n_column, token_GLenum, n_enum);
		return true;
	}
};

class CMaterialLexer::CEmitCData : public CTokenEmit {
public:
	CEmitCData(TToken &r_t_out_token, bool &r_b_validity_flag)
		:CTokenEmit(r_t_out_token, r_b_validity_flag)
	{}

	virtual bool operator ()(const char *p_s_buffer,
		int n_regexp_id, int n_line, int n_column) const
	{
		_ASSERTE(n_regexp_id == 7);
		_ASSERTE(strstr(p_s_buffer, "<![CDATA[") == p_s_buffer);

		m_r_b_validity_flag = true;
		p_s_buffer += 9; // strlen("<![CDATA[");

		int n_length = strlen(p_s_buffer);
		char *p_s_string;
		if(!(p_s_string = new char[n_length - 2]))
			return false;
		strncpy(p_s_string, p_s_buffer, n_length - 3);
		p_s_string[n_length - 3] = 0;
		// copy <![CDATA[string]]>

		m_r_t_out_token = TToken(n_line, n_column, token_CData, p_s_string);
		delete[] p_s_string;
		return m_r_t_out_token.p_s_Value() != 0;
	}
};

/*
 *								=== ~CMaterialLexer::CTokenEmit ===
 */

/*
 *								=== CMaterialLexer ===
 */

/*const CMaterialLexer::TDrivingTable::TTransition CMaterialLexer::m_p_transition[77] = {
    {9, 10, 1}, {13, 13, 1}, {32, 32, 1}, {40, 41, 4}, {58, 59, 4}, {123, 123, 4},
    {125, 125, 4}, {47, 47, 7}, {39, 39, 3}, {46, 46, 6}, {65, 90, 10}, {43, 43, 5},
    {45, 45, 5}, {97, 122, 11}, {34, 34, 2}, {48, 57, 8}, {60, 60, 9}, {9, 10, 1},
    {13, 13, 1}, {32, 32, 1}, {0, 33, 2}, {35, 91, 2}, {93, 255, 2}, {34, 34, 31},
    {92, 92, 32}, {0, 38, 3}, {40, 255, 3}, {39, 39, 31}, {46, 46, 6}, {48, 57, 8},
    {48, 57, 26}, {42, 42, 23}, {46, 46, 6}, {48, 57, 8}, {33, 33, 12}, {65, 90, 10},
    {95, 95, 10}, {45, 45, 11}, {48, 57, 11}, {97, 122, 11}, {91, 91, 13}, {67, 67, 14},
    {68, 68, 15}, {65, 65, 16}, {84, 84, 17}, {65, 65, 18}, {91, 91, 19}, {93, 93, 20},
    {0, 92, 19}, {94, 255, 19}, {0, 92, 19}, {94, 255, 19}, {93, 93, 21}, {0, 61, 19},
    {63, 255, 19}, {62, 62, 22}, {0, 41, 23}, {43, 255, 23}, {42, 42, 24}, {0, 46, 23},
    {48, 255, 23}, {47, 47, 25}, {102, 102, 28}, {101, 101, 27}, {48, 57, 26}, {43, 43, 29},
    {45, 45, 29}, {48, 57, 30}, {48, 57, 30}, {102, 102, 28}, {48, 57, 30}, {34, 34, 2},
    {39, 39, 2}, {92, 92, 2}, {110, 110, 2}, {114, 114, 2}, {116, 116, 2}
};

CMaterialLexer::TDrivingTable::TState CMaterialLexer::m_p_state[33] = {
	{ 0 + m_p_transition, -1, 17}, { 17 + m_p_transition,0, 3},
	{20 + m_p_transition, -1, 5}, { 25 + m_p_transition,-1, 3},
	{28 + m_p_transition, 2, 0}, { 28 + m_p_transition,-1, 2},
	{30 + m_p_transition, -1, 1}, { 31 + m_p_transition,-1, 1},
	{32 + m_p_transition, 4, 2}, { 34 + m_p_transition,-1, 1},
	{35 + m_p_transition, 6, 2}, { 37 + m_p_transition,1, 3},
	{40 + m_p_transition, -1, 1}, { 41 + m_p_transition,-1, 1},
	{42 + m_p_transition, -1, 1}, { 43 + m_p_transition,-1, 1},
	{44 + m_p_transition, -1, 1}, { 45 + m_p_transition,-1, 1},
	{46 + m_p_transition, -1, 1}, { 47 + m_p_transition,-1, 3},
	{50 + m_p_transition, -1, 3}, { 53 + m_p_transition,-1, 3},
	{56 + m_p_transition, 7, 0}, { 56 + m_p_transition,-1, 3},
	{59 + m_p_transition, -1, 3}, { 62 + m_p_transition,0, 0},
	{62 + m_p_transition, 5, 3}, { 65 + m_p_transition,-1, 3},
	{68 + m_p_transition, 5, 0}, { 68 + m_p_transition,-1, 1},
	{69 + m_p_transition, 5, 2}, { 71 + m_p_transition,3, 0},
	{71 + m_p_transition, -1, 6}
};*/

const CMaterialLexer::TDrivingTable::TTransition CMaterialLexer::m_p_transition[78] = {
	{9, 10, 1}, {13, 13, 1}, {32, 32, 1}, {46, 46, 6}, {39, 39, 3}, {43, 43, 5},
	{45, 45, 5}, {40, 41, 4}, {58, 59, 4}, {123, 123, 4}, {125, 125, 4}, {47, 47, 7},
	{48, 57, 8}, {97, 122, 11}, {60, 60, 9}, {65, 90, 10}, {34, 34, 2}, {9, 10, 1},
	{13, 13, 1}, {32, 32, 1}, {0, 33, 2}, {35, 91, 2}, {93, 255, 2}, {34, 34, 31},
	{92, 92, 32}, {0, 38, 3}, {40, 255, 3}, {39, 39, 31}, {46, 46, 6}, {48, 57, 8},
	{48, 57, 30}, {42, 42, 27}, {46, 46, 6}, {48, 57, 8}, {101, 101, 23}, {33, 33, 12},
	{65, 90, 10}, {95, 95, 10}, {45, 45, 11}, {48, 57, 11}, {97, 122, 11}, {91, 91, 13},
	{67, 67, 14}, {68, 68, 15}, {65, 65, 16}, {84, 84, 17}, {65, 65, 18}, {91, 91, 19},
	{0, 92, 19}, {94, 255, 19}, {93, 93, 20}, {0, 92, 19}, {94, 255, 19}, {93, 93, 21},
	{0, 61, 19}, {63, 255, 19}, {62, 62, 22}, {43, 43, 24}, {45, 45, 24}, {48, 57, 25},
	{48, 57, 25}, {48, 57, 25}, {102, 102, 26}, {0, 41, 27}, {43, 255, 27}, {42, 42, 28},
	{0, 46, 27}, {48, 255, 27}, {47, 47, 29}, {102, 102, 26}, {101, 101, 23}, {48, 57, 30},
	{34, 34, 2}, {39, 39, 2}, {92, 92, 2}, {110, 110, 2}, {114, 114, 2}, {116, 116, 2}
};

CMaterialLexer::TDrivingTable::TState CMaterialLexer::m_p_state[33] = {
	{m_p_transition + 0, -1, 17}, {m_p_transition + 17, 0, 3},
	{m_p_transition + 20, -1, 5}, {m_p_transition + 25, -1, 3},
	{m_p_transition + 28, 2, 0}, {m_p_transition + 28, -1, 2},
	{m_p_transition + 30, -1, 1}, {m_p_transition + 31, -1, 1},
	{m_p_transition + 32, 4, 3}, {m_p_transition + 35, -1, 1},
	{m_p_transition + 36, 6, 2}, {m_p_transition + 38, 1, 3},
	{m_p_transition + 41, -1, 1}, {m_p_transition + 42, -1, 1},
	{m_p_transition + 43, -1, 1}, {m_p_transition + 44, -1, 1},
	{m_p_transition + 45, -1, 1}, {m_p_transition + 46, -1, 1},
	{m_p_transition + 47, -1, 1}, {m_p_transition + 48, -1, 3},
	{m_p_transition + 51, -1, 3}, {m_p_transition + 54, -1, 3},
	{m_p_transition + 57, 7, 0}, {m_p_transition + 57, -1, 3},
	{m_p_transition + 60, -1, 1}, {m_p_transition + 61, 5, 2},
	{m_p_transition + 63, 5, 0}, {m_p_transition + 63, -1, 3},
	{m_p_transition + 66, -1, 3}, {m_p_transition + 69, 0, 0},
	{m_p_transition + 69, 5, 3}, {m_p_transition + 72, 3, 0},
	{m_p_transition + 72, -1, 6}
};
// new tables with fixed float exponent notation

/*
 *	CMaterialLexer::CMaterialLexer(const char *p_s_filename)
 *		- default constructor; creates a new parser, bound to p_s_filename
 *		- call b_Status() to see wheter file was opened successfully
 *		  (no parsing occurs at this moment)
 *		- note all parsing occurs in us-ascii encoding
 */
CMaterialLexer::CMaterialLexer(const char *p_s_filename)
	:m_t_token(0, 0, token_EOF), m_scanner(m_p_state, m_p_emit, 8)
{
	if((m_p_emit[0] = new CEmitSkip(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[1] = new CEmitNode(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[2] = new CEmitOperator(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[3] = new CEmitString(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[4] = new CEmitInt(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[5] = new CEmitFloat(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[6] = new CEmitGLenum(m_t_token, m_b_token_valid)) &&
	   (m_p_emit[7] = new CEmitCData(m_t_token, m_b_token_valid))) {
		if((m_p_file = new CFileReader(p_s_filename)) && !m_p_file->b_Status()) {
			delete m_p_file;
			m_p_file = 0;
		}
	} else
		m_p_file = 0;
	// init token emit objects, create file reader and open given file
}

/*
 *	CMaterialLexer::~CMaterialLexer()
 *		- default destructor
 */
CMaterialLexer::~CMaterialLexer()
{
	for(int i = 0; i < 8; ++ i) {
		if(m_p_emit[i])
			delete m_p_emit[i];
	}
	if(m_p_file)
		delete m_p_file;
}

/*
 *	bool CMaterialLexer::b_Status() const
 *		- returns status of parser
 *		- returns true if no errors occured, returns false if
 *		  constructor or parsing failed
 */
bool CMaterialLexer::b_Status() const
{
	return m_p_file != 0;
}

/*
 *	bool CMaterialLexer::Get_NextToken()
 *		- parses next token from file
 *		- returns true on success, false on parsing / memory failure
 *		- note reading past the end of the file is not an error
 *		- always fails in case b_Status() returns false
 */
bool CMaterialLexer::Get_NextToken()
{
	if(!m_p_file)
		return false;

	for(;;) {
		m_b_token_valid = false;
		if(!m_scanner.GetToken(*m_p_file))
			return false;
		if(!m_b_token_valid) {
			m_t_token = TToken(m_scanner.n_Cur_Line(),
				m_scanner.n_Cur_Column(), token_EOF);
			return true;
		} else if(m_t_token.n_Type() != token_White)
			return true;
	}
	// read token which is not whitespace
}

/*
 *	int CMaterialLexer::n_Cur_Line() const
 *		- returns zero-based index of line where the parser last stopped
 *		- this is useful for printing error messages on where parsing failed
 */
int CMaterialLexer::n_Cur_Line() const
{
	return m_scanner.n_Cur_Line();
}

/*
 *	int CMaterialLexer::n_Cur_Column() const
 *		- returns zero-based index of column where the parser last stopped
 *		- this is useful for printing error messages on where parsing failed
 */
int CMaterialLexer::n_Cur_Column() const
{
	return m_scanner.n_Cur_Column();
}

/*
 *	const CMaterialLexer::TToken &CMaterialLexer::t_Cur_Token() const
 *		- returns constant reference to current token
 *		- note in case no token was read, this returns token_EOF at line 0, col 0
 *		- return value is undefined in case b_Status() returns false
 */
const CMaterialLexer::TToken &CMaterialLexer::t_Cur_Token() const
{
	return m_t_token;
}

/*
 *								=== ~CMaterialLexer ===
 */

/*
 *								=== CGLEnumTable ===
 */

const CGLenumTable::TTable CGLenumTable::m_p_table[] = {
	{"GL_POINT", 1/*GL_POINT*/}/*,
	{"GL_LINE", GL_LINE},
	{"GL_FILL", GL_FILL},

	{"GL_FRONT", GL_FRONT},
	{"GL_BACK", GL_BACK},
	{"GL_FRONT_AND_BACK", GL_FRONT_AND_BACK},

	{"GL_CW", GL_CW},
	{"GL_CCW", GL_CCW},

	{"GL_ZERO", GL_ZERO},
	{"GL_ONE", GL_ONE},
	{"GL_SRC_COLOR", GL_SRC_COLOR},
	{"GL_ONE_MINUS_SRC_COLOR", GL_ONE_MINUS_SRC_COLOR},
	{"GL_DST_COLOR", GL_DST_COLOR},
	{"GL_ONE_MINUS_DST_COLOR", GL_ONE_MINUS_DST_COLOR},
	{"GL_SRC_ALPHA", GL_SRC_ALPHA},
	{"GL_ONE_MINUS_SRC_ALPHA", GL_ONE_MINUS_SRC_ALPHA},
	{"GL_DST_ALPHA", GL_DST_ALPHA},
	{"GL_ONE_MINUS_DST_ALPHA", GL_ONE_MINUS_DST_ALPHA},
	{"GL_CONSTANT_COLOR", GL_CONSTANT_COLOR},
	{"GL_ONE_MINUS_CONSTANT_COLOR", GL_ONE_MINUS_CONSTANT_COLOR},
	{"GL_CONSTANT_ALPHA", GL_CONSTANT_ALPHA},
	{"GL_ONE_MINUS_CONSTANT_ALPHA", GL_ONE_MINUS_CONSTANT_ALPHA},
	{"GL_SRC_ALPHA_SATURATE", GL_SRC_ALPHA_SATURATE}, // only valid for src color / alpha

	{"GL_FUNC_ADD", GL_FUNC_ADD},
	{"GL_FUNC_SUBTRACT", GL_FUNC_SUBTRACT},
	{"GL_FUNC_REVERSE_SUBTRACT", GL_FUNC_REVERSE_SUBTRACT},
	{"GL_MIN", GL_MIN},
	{"GL_MAX", GL_MAX},

	{"GL_NEVER", GL_NEVER},
	{"GL_LESS", GL_LESS},
	{"GL_EQUAL", GL_EQUAL},
	{"GL_LEQUAL", GL_LEQUAL},
	{"GL_GREATER", GL_GREATER},
	{"GL_NOTEQUAL", GL_NOTEQUAL},
	{"GL_GEQUAL", GL_GEQUAL},
	{"GL_ALWAYS", GL_ALWAYS},

	{"GL_KEEP", GL_KEEP},
	{"GL_ZERO", GL_ZERO},
	{"GL_REPLACE", GL_REPLACE},
	{"GL_INCR", GL_INCR},
	{"GL_DECR", GL_DECR},
	{"GL_INVERT", GL_INVERT},

	{"GL_RGBA", GL_RGBA},
	{"GL_RGB", GL_RGB},
	{"GL_ALPHA", GL_ALPHA},
	{"GL_INTENSITY", GL_INTENSITY},
	{"GL_LUMINANCE", GL_LUMINANCE},
	{"GL_LUMINANCE_ALPHA", GL_LUMINANCE_ALPHA},
	{"GL_RGBA32F", GL_RGBA32F_ARB},
	{"GL_RGB32F", GL_RGB32F_ARB},
	{"GL_ALPHA32F", GL_ALPHA32F_ARB},
	{"GL_INTENSITY32F", GL_INTENSITY32F_ARB},
	{"GL_LUMINANCE32F", GL_LUMINANCE32F_ARB},
	{"GL_LUMINANCE_ALPHA32F", GL_LUMINANCE_ALPHA32F_ARB},
	{"GL_RGBA16F", GL_RGBA16F_ARB},
	{"GL_RGB16F", GL_RGB16F_ARB},
	{"GL_ALPHA16F", GL_ALPHA16F_ARB},
	{"GL_INTENSITY16F", GL_INTENSITY16F_ARB},
	{"GL_LUMINANCE16F", GL_LUMINANCE16F_ARB},
	{"GL_LUMINANCE_ALPHA16F", GL_LUMINANCE_ALPHA16F_ARB},
	{"GL_DEPTH_COMPONENT16", GL_DEPTH_COMPONENT16},
	{"GL_DEPTH_COMPONENT24", GL_DEPTH_COMPONENT24},
	{"GL_DEPTH_COMPONENT32", GL_DEPTH_COMPONENT32},
	// stencil component not allowed as texture format

	{"GL_CLAMP", GL_CLAMP},
	{"GL_CLAMP_TO_EDGE", GL_CLAMP_TO_EDGE},
	{"GL_REPEAT", GL_REPEAT},
	{"GL_CLAMP_TO_BORDER", GL_CLAMP_TO_BORDER},
	{"GL_MIRRORED_REPEAT", GL_MIRRORED_REPEAT},

	{"GL_NEAREST", GL_NEAREST},
	{"GL_LINEAR", GL_LINEAR},
	{"GL_NEAREST_MIPMAP_NEAREST", GL_NEAREST_MIPMAP_NEAREST},
	{"GL_NEAREST_MIPMAP_LINEAR", GL_NEAREST_MIPMAP_LINEAR},
	{"GL_LINEAR_MIPMAP_NEAREST", GL_LINEAR_MIPMAP_NEAREST},
	{"GL_LINEAR_MIPMAP_LINEAR", GL_LINEAR_MIPMAP_LINEAR},

	{"GL_NONE", GL_NONE},
	{"GL_COMPARE_R_TO_TEXTURE", GL_COMPARE_R_TO_TEXTURE}*/
};

/*
 *	static int CGLenumTable::n_Value(const char *p_s_name)
 *		- returns value of OpenGL enum p_s_name
 *		- returns -1 in case p_s_name is not found in the table
 */
int CGLenumTable::n_Value(const char *p_s_name)
{
	/*if(strncmp(p_s_name, "GL_", 2))
		return -1;
	p_s_name += 3;*/
	for(unsigned int i = 0; i < sizeof(m_p_table) / sizeof(m_p_table[0]); ++ i) {
		if(!strcmp(p_s_name, m_p_table[i].p_s_name))
			return m_p_table[i].n_value;
	}
	return -1;
}

/*
 *								=== ~CGLEnumTable ===
 */
