/*
								+----------------------------------+
								|                                  |
								| ***  Random generator model  *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|           RandGen.cpp            |
								|                                  |
								+----------------------------------+
*/

#include <stdlib.h>
#include "RandGen.h"

/*
 *								=== CStdLibGenerator ===
 */

/*
 *	CStdLibGenerator::CStdLibGenerator(int n_initial_seed = 0)
 *		- default constructor, sets initial seed
 *		- note there shouldn't be multiple instances of this class working
 *		  simultaneously because stdlib's rand() is global and thus shared
 */
CStdLibGenerator::CStdLibGenerator(int n_initial_seed)
{
	SetSeed(n_initial_seed);
}

/*
 *	virtual unsigned int CStdLibGenerator::n_Seed() const
 *		- returns seed generator was initialized with
 */
unsigned int CStdLibGenerator::n_Seed() const
{
	return m_n_seed;
}

/*
 *	virtual void CStdLibGenerator::SetSeed(unsigned int n_seed)
 *		- sets new seed n_seed and resets generator
 */
void CStdLibGenerator::SetSeed(unsigned int n_seed)
{
	srand(m_n_seed = n_seed);
}

/*
 *	virtual void CStdLibGenerator::Reset()
 *		- resets generator while not changing seed
 *		  (ie. the same number sequence will be generated again)
 */
void CStdLibGenerator::Reset()
{
	srand(m_n_seed);
}

/*
 *	virtual float CStdLibGenerator::f_Rand()
 *		- returns floating-point random number in range 0 to 1 inclusive
 */
float CStdLibGenerator::f_Rand()
{
	return rand() / (float)RAND_MAX;
}

/*
 *								=== ~CStdLibGenerator ===
 */
