/*
								+----------------------------------+
								|                                  |
								| ***  Random generator model  *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|            RandGen.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __RANDOM_GENERATOR_INCLUDED
#define __RANDOM_GENERATOR_INCLUDED

/*
 *	class CMinRandomGenerator
 *		- generic minimal random number generator class (intended
 *		  for experiments with random number generation methods)
 */
class CMinRandomGenerator {
public:
	/*
	 *	virtual float CRandomGenerator::f_Rand()
	 *		- returns floating-point random number in range 0 to 1 inclusive
	 */
	virtual float f_Rand() = 0;
};

/*
 *	class CRandomGenerator
 *		- generic random number generator class (intended
 *		  for experiments with random number generation methods)
 */
class CRandomGenerator : public CMinRandomGenerator {
public:
	/*
	 *	virtual unsigned int CRandomGenerator::n_Seed() const
	 *		- returns seed generator was initialized with
	 */
	virtual unsigned int n_Seed() const = 0;

	/*
	 *	virtual void CRandomGenerator::SetSeed(unsigned int n_seed)
	 *		- sets new seed n_seed and resets generator
	 */
	virtual void SetSeed(unsigned int n_seed) = 0;

	/*
	 *	virtual void CRandomGenerator::Reset()
	 *		- resets generator while not changing seed
	 *		  (ie. the same number sequence will be generated again)
	 */
	virtual void Reset() = 0;
};

/*
 *	class CStdLibGenerator
 *		- stdlib's rand() based random number generator
 */
class CStdLibGenerator : public CRandomGenerator {
protected:
	unsigned int m_n_seed;

public:
	/*
	 *	CStdLibGenerator::CStdLibGenerator(int n_initial_seed = 0)
	 *		- default constructor, sets initial seed
	 *		- note there shouldn't be multiple instances of this class working
	 *		  simultaneously because stdlib's rand() is global and thus shared
	 */
	CStdLibGenerator(int n_initial_seed = 0);

	/*
	 *	virtual unsigned int CStdLibGenerator::n_Seed() const
	 *		- returns seed generator was initialized with
	 */
	virtual unsigned int n_Seed() const;

	/*
	 *	virtual void CStdLibGenerator::SetSeed(unsigned int n_seed)
	 *		- sets new seed n_seed and resets generator
	 */
	virtual void SetSeed(unsigned int n_seed);

	/*
	 *	virtual void CStdLibGenerator::Reset()
	 *		- resets generator while not changing seed
	 *		  (ie. the same number sequence will be generated again)
	 */
	virtual void Reset();

	/*
	 *	virtual float CStdLibGenerator::f_Rand()
	 *		- returns floating-point random number in range 0 to 1 inclusive
	 */
	virtual float f_Rand();
};

#endif //__RANDOM_GENERATOR_INCLUDED
