/*
								+----------------------------------+
								|                                  |
								|     ***   Targa writer   ***     |
								|                                  |
								|    Design by Lukas Polok 2007    |
								|    Implemented by Vit Kolar      |
								|                                  |
								|          TgaOutput.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <string.h>
#include <stdio.h>
#include "TgaOutput.h"

#pragma pack(1)

struct TTgaHeader {
	unsigned char n_id_length;
	unsigned char n_palette_type;
	unsigned char n_image_type;
	unsigned short n_first_color;
	unsigned short n_palette_colors;
	unsigned char n_palette_entry_size;
	unsigned short n_left;
	unsigned short n_top;
	unsigned short n_image_width;
	unsigned short n_image_height;
	unsigned char n_bpp;
	unsigned char n_descriptor_bits;
};

#pragma pack()

/*
 *	bool Save_TGA(const char *p_s_file, int n_width, int n_height,
 *		const unsigned int *p_buffer)
 *		- write RGB targa image to file p_s_file
 *		- returns true on success, false on failure
 *
 *		- zadn:
 *			- tahle funkce m uloit RGB 24 bpp targu (tedy bez alfa-kanlu) o rozmrech
 *			  n_width x n_height do souboru s nzvem p_s_file (jednodue pepsat pokud existuje)
 *			- p_buffer obsahuje jednotliv pixely bitmapy, jak jdou po dcch, zdola nahoru.
 *			- kad pixel m hodnotu (r << 16) + (g << 8) + b, kde r, g, b jsou unsigned char
 *			- vrac se true pi spchu, false pi chyb dn hlen se nevypisuje)
 */
bool Save_TGA(const char *p_s_file, int n_width, int n_height, const unsigned int *p_buffer)
{
	FILE *p_fw;
	if(!(p_fw = fopen(p_s_file, "wb")))
		return false;

	TTgaHeader t_header;
    memset(&t_header, 0, sizeof(TTgaHeader));
	//t_header.n_id_length = 0;
	t_header.n_bpp = 24;
	t_header.n_image_type = 2; // rgb
	//t_header.n_palette_type = 0;
	//t_header.n_palette_entry_size = 8;
	//t_header.n_first_color = 0;
	//t_header.n_palette_colors = 0;
	//t_header.n_left = 0;
	//t_header.n_top = 0;
	t_header.n_image_width = n_width;
	t_header.n_image_height = n_height;
	t_header.n_descriptor_bits = 0x20; // upside-down
	// create header

	if(fwrite(&t_header, sizeof(t_header), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write header

	for(const unsigned int *p_scanline = p_buffer, *p_end = p_buffer +
	   (n_width * n_height); p_scanline != p_end; p_scanline += n_width) {
		for(const unsigned int *p_ptr = p_scanline, *p_end2 = p_scanline + n_width;
		   p_ptr != p_end2; ++ p_ptr) {
			unsigned int n_rgba = *p_ptr;
			unsigned char p_data[3] = {n_rgba >> 16, n_rgba >> 8, n_rgba};
			if(fwrite(p_data, sizeof(char), 3, p_fw) != 3) {
				fclose(p_fw);
				return false;
			}
		}
	}
	// write data

	fclose(p_fw);

	return true;
}

/*
 *		-end-of-file-
 */
